# -*- coding: utf-8 -*-

# built ins
import asyncio

# installed
from loguru import logger as log

# user defined formulas
from streaming_helper.db_management import sqlite_management as db_mgt
from streaming_helper.restful_api.deribit import end_point_params_template as end_point
from streaming_helper.restful_api import connector
from streaming_helper.utilities import (
    error_handling,
    pickling,
    string_modification as str_mod,
    system_tools,
    time_modification as time_mod,
)


async def initial_procedures(
    private_data: object,
    client_redis: object,
    config_app: list,
) -> None:

    try:

        # get tradable strategies
        tradable_config_app = config_app["tradable"]

        # get TRADABLE currencies
        currencies: list = [o["spot"] for o in tradable_config_app][0]

        basic_https_connection_url = end_point.basic_https()

        endpoint_currencies = end_point.get_currencies_end_point()

        # get ALL traded currencies in deribit
        get_currencies_all = await connector.get_connected(
            basic_https_connection_url,
            endpoint_currencies,
        )

        all_exc_currencies = [o["currency"] for o in get_currencies_all["result"]]

        server_time = time_mod.get_now_unix_time()

        ONE_SECOND = 1000

        one_minute = ONE_SECOND * 60

        five_days_ago = server_time - (one_minute * 60 * 24 * 5)

        my_path_cur = system_tools.provide_path_for_file("currencies")

        pickling.replace_data(
            my_path_cur,
            all_exc_currencies,
        )

        for currency in all_exc_currencies:

            get_instruments_end_point = end_point.get_instruments(currency)

            endpoint_instruments = end_point.get_instruments_end_point(currency)

            # get ALL traded currencies in deribit
            instruments = await connector.get_connected(
                basic_https_connection_url,
                endpoint_instruments,
            )
            my_path_instruments = system_tools.provide_path_for_file(
                "instruments", currency
            )

            pickling.replace_data(
                my_path_instruments,
                instruments,
            )

        for currency in currencies:

            currency_lower = currency.lower()

            archive_db_table = f"my_trades_all_{currency_lower}_json"

            query_trades_active_basic = f"SELECT instrument_name, user_seq, timestamp, trade_id  FROM  {archive_db_table}"

            query_trades_active_where = f"WHERE instrument_name LIKE '%{currency}%'"

            query_trades = f"{query_trades_active_basic} {query_trades_active_where}"

            my_trades_currency = await db_mgt.executing_query_with_return(query_trades)

            if my_trades_currency == []:

                await refill_db(
                    private_data,
                    archive_db_table,
                    currency,
                    five_days_ago,
                )

    except Exception as error:

        await error_handling.parse_error_message_with_redis(
            client_redis,
            error,
        )


async def refill_db(
    private_data: object,
    archive_db_table: str,
    currency: str,
    five_days_ago: int,
) -> None:

    transaction_log = await private_data.get_transaction_log(
        currency,
        five_days_ago,
        1000,
        "trade",
    )

    await distributing_transaction_log_from_exchange(
        archive_db_table,
        transaction_log,
    )


async def distributing_transaction_log_from_exchange(
    archive_db_table: str,
    transaction_log: list,
) -> None:

    log.warning(f"transaction_log {transaction_log}")

    if transaction_log:

        for transaction in transaction_log:
            result = {}

            if "sell" in transaction["side"]:
                direction = "sell"

            if "buy" in transaction["side"]:
                direction = "buy"

            result.update({"trade_id": transaction["trade_id"]})
            result.update({"user_seq": transaction["user_seq"]})
            result.update({"side": transaction["side"]})
            result.update({"timestamp": transaction["timestamp"]})
            result.update({"position": transaction["position"]})
            result.update({"amount": transaction["amount"]})
            result.update({"order_id": transaction["order_id"]})
            result.update({"price": transaction["price"]})
            result.update({"instrument_name": transaction["instrument_name"]})
            result.update({"label": None})
            result.update({"direction": direction})
            result.update({"currency": transaction["currency"]})

            await db_mgt.insert_tables(
                archive_db_table,
                result,
            )


def portfolio_combining(
    portfolio_all: list,
    portfolio_channel: str,
    result_template: dict,
) -> dict:

    portfolio = (
        [o["portfolio"] for o in portfolio_all if o["type"] == "subaccount"][0]
    ).values()

    #! need to update currency to upper

    result_template["params"].update({"data": portfolio})
    result_template["params"].update({"channel": portfolio_channel})

    return result_template


def my_trades_active_combining(
    my_trades_active_all: list,
    my_trades_channel: str,
    result_template: dict,
) -> dict:

    result_template["params"].update({"data": my_trades_active_all})
    result_template["params"].update({"channel": my_trades_channel})

    return result_template


def sub_account_combining(
    sub_accounts: list,
    sub_account_cached_channel: str,
    result_template: dict,
) -> dict:

    orders_cached = []
    positions_cached = []

    try:

        for sub_account in sub_accounts:

            sub_account = sub_account[0]

            sub_account_orders = sub_account["open_orders"]

            if sub_account_orders:

                for order in sub_account_orders:

                    orders_cached.append(order)

            sub_account_positions = sub_account["positions"]

            if sub_account_positions:

                for position in sub_account_positions:

                    positions_cached.append(position)

        sub_account = dict(
            orders_cached=orders_cached,
            positions_cached=positions_cached,
        )

        result_template["params"].update({"data": sub_account})
        result_template["params"].update({"channel": sub_account_cached_channel})

        return result_template

    except:

        sub_account = dict(
            orders_cached=[],
            positions_cached=[],
        )

        result_template["params"].update({"data": sub_account})
        result_template["params"].update({"channel": sub_account_cached_channel})

        return result_template


def is_order_allowed_combining(
    all_instruments_name: list,
    order_allowed_channel: str,
    result_template: dict,
) -> dict:

    combined_order_allowed = []
    for instrument_name in all_instruments_name:

        currency: str = str_mod.extract_currency_from_text(instrument_name)

        if "-FS-" in instrument_name:
            size_is_reconciled = 1

        else:
            size_is_reconciled = 0

        order_allowed = dict(
            instrument_name=instrument_name,
            size_is_reconciled=size_is_reconciled,
            currency=currency,
        )

        combined_order_allowed.append(order_allowed)

    result_template["params"].update({"data": combined_order_allowed})
    result_template["params"].update({"channel": order_allowed_channel})

    return result_template



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
https://medium.com/@danielbriano

Binance Asynchronous OHLCV Downloader

This module provides a class to download OHLCV (Open, High, Low, Close, Volume) data
from the Binance API as efficiently as possible using asynchronous requests.

Key Features:
- Asynchronous data fetching for improved performance
- Rate limiting to comply with Binance API restrictions
- Circuit breaker pattern to handle and recover from API errors
- Adjustable request rates based on API usage feedback

Classes:
- BinanceClient: Main class for interacting with the Binance API
- RateLimiter: Manages and adjusts request rates
- CircuitBreaker: Implements the circuit breaker pattern for error handling
- RateLimitManager: Manages rate limit states and ban durations

Usage:
    client = BinanceClient()
    ohlcv_data = await client.get_ohlcv('BTCUSDT', '1h', start_time, end_time)

Note:
This module is designed to work within the constraints of the Binance API's rate limits.
Users should be aware of and respect these limits to avoid temporary IP bans.

Binance API Rate Limits:
- REQUEST_WEIGHT: 1200 per minute
- ORDERS: 50 per 10 seconds, 160000 per day
- RAW_REQUESTS: 6100 per 5 minutes

Created on Dec 17 06:36:20 2024

@author: dhaneor
"""
import asyncio
import logging
import random
from dataclasses import dataclass
from datetime import datetime, timedelta
from threading import Lock
from time import time
from typing import Any, Dict, List, Tuple

import aiohttp

LOG_LEVEL = logging.INFO

# Increase this for increased speed at the beginning of a download/minute.
# 15 seems to be a good number which means that at the end of the minute,
# there is only a short waiting time for the reset of the API 1m limit.
# If the latency is low, it may even make sense to decrease this value.
MAX_WORKERS = 15  # max number of parallel requests to the Binance API
MAX_RETRIES = 5  # number of retries before giving up
# These are the default values that can also be overridden when initializing
# the BinanceClient class.

BASE_URL = "https://api.binance.com"
API_ENDPOINT = "/api/v3/klines"

KLINES_LIMIT = 1000  # how may klines to request in one API call
BASE_DELAY = MAX_WORKERS / 25  # base delay for rate limiting (in seconds)

RATE_LIMIT_BARRIER = 1200  # this is the Binance API rate limit (see above)
HARD_LIMIT = RATE_LIMIT_BARRIER - 50  # limit hard above this threshhold
SOFT_LIMIT = RATE_LIMIT_BARRIER * 0.6  # start limiting above this threshhold

RAW_REQUEST_LIMIT = 6100  # requests per 5 minutes (see above)
LONG_DELAY_THRESHHOLD = RAW_REQUEST_LIMIT * 0.9  # add even more delay above ...

SIMULATE_429_ERRORS = False  # activate this to test the RateLimitManager

if __name__ == "__main__":
    logger = logging.getLogger(__name__)
    logger.setLevel(LOG_LEVEL)

    ch = logging.StreamHandler()
    ch.setLevel(LOG_LEVEL)

    formatter = logging.Formatter(
        "%(asctime)s - %(name)s.%(funcName)s.%(lineno)d  - [%(levelname)s]: %(message)s"
    )
    ch.setFormatter(formatter)

    logger.addHandler(ch)
else:
    logger = logging.getLogger(f"main.{__name__}")
    logger.setLevel(LOG_LEVEL)


class RateLimitManager:
    """Class to manage 429 and 418 errors."""

    def __init__(self):
        self._lock = Lock()
        self._retry_after = 0
        self._ban_until = 0

    def get_wait_time(self) -> float:
        current_time = time()
        if current_time < self._ban_until:
            return self._ban_until - current_time
        if current_time < self._retry_after:
            return self._retry_after - current_time
        return 0

    def set_rate_limit(self, retry_after):
        self._retry_after = time() + retry_after

    def set_ban(self, ban_duration):
        self._ban_until = time() + ban_duration


class CircuitBreaker:
    """Circuit Breaker class."""

    def __init__(self, max_failures=5, reset_time=300):
        self.failures = 0
        self.max_failures = max_failures
        self.reset_time = reset_time
        self.last_failure_time = None

    def record_failure(self) -> None:
        current_time = datetime.now()

        if self.last_failure_time:
            secs_since_reset = timedelta(seconds=self.reset_time)
            if (current_time - self.last_failure_time) > secs_since_reset:
                self.failures = 0

        self.failures += 1
        self.last_failure_time = current_time

    def is_open(self) -> bool:
        return self.failures >= self.max_failures

    def seconds_to_reset(self) -> float:
        if self.is_open():
            return (self.last_failure_time + self.reset_time) - time()
        else:
            return 0


@dataclass
class RateLimiter:
    """Rate limiter class"""

    weight_1m: int = 0
    weight_total: int = 0
    last_update: float = time()
    time_offset: float = 0.0
    throttle_counter: int = 0

    def update(self, weight_1m: int, weight_total: int):
        # reset at the start of each minute
        if self.seconds_to_next_full_minute() > 55:
            self.weight_1m = 0
            self.weight_total = 0

        # update values
        self.weight_1m = max(weight_1m, self.weight_1m)
        self.weight_total = max(weight_total, self.weight_total)
        self.last_update = time()

    def should_throttle(self) -> bool:
        return self.weight_1m > SOFT_LIMIT or self.weight_total > LONG_DELAY_THRESHHOLD

    def chill_out_bro(self) -> tuple[float, int]:
        # calculate the delay we need to prevent 'too many requests'
        seconds_left = self.seconds_to_next_full_minute() + 0.5

        # start throttling when the 1m weight goes above the soft limit
        base_delay = BASE_DELAY
        short_delay = 0.0

        if self.weight_1m >= SOFT_LIMIT:
            short_delay = base_delay + (
                seconds_left**2 / (RATE_LIMIT_BARRIER - (self.weight_1m - 1))
            )

        # wait for the next minute before proceeding when we approach
        # the API limit of 1200 used weight per one minute
        # but do not wait longer than 55 seconds (even though we try to sync to
        # to the server time, sometimes the delay was set to 59 or 60 seconds)
        if self.weight_1m >= HARD_LIMIT:
            short_delay = seconds_left if seconds_left < 55 else 5

        # handle the Raw Request Limit ... usually, this should never happen,
        # because if the above code works as intended then we will not reach
        # more than 6000 requests per 5 minutes
        add_delay = 0.0

        if self.weight_total > LONG_DELAY_THRESHHOLD:
            logger.info(f"increased total weight: {self.weight_total}")
            add_delay = (self.weight_total - LONG_DELAY_THRESHHOLD) / 50

        # add even more delay when we have used up more than 95%
        # of the weight per 5 minutes
        if self.weight_total > RAW_REQUEST_LIMIT * 0.95:
            logger.warning("approaching 5m limit ... increasing delay")
            base_delay = 10

        self.throttle_counter += 1  # just for logging purposes
        return short_delay + add_delay, self.throttle_counter

    def seconds_to_next_full_minute(self) -> float:
        adjusted_time = int(time() + self.time_offset)
        now = datetime.fromtimestamp(adjusted_time)
        return 60 - now.second


class BinanceClient:
    """Specialized Binance client for OHLCV data downloads."""

    def __init__(
        self,
        max_workers: int | None = None,
        max_retries: int | None = None,
        timeout: int = 10,
    ) -> None:
        self.max_workers = max_workers or MAX_WORKERS
        self.max_retries = max_retries or MAX_RETRIES
        self.timeout = aiohttp.ClientTimeout(total=timeout)

        self.rate_limiter = RateLimiter()
        self.circuit_breaker = CircuitBreaker()
        self.rate_limit_manager = RateLimitManager()
        self.semaphore = asyncio.Semaphore(self.max_workers)

    async def sync_server_time(self):
        async with aiohttp.ClientSession() as session:
            async with session.get(f"{BASE_URL}/api/v3/time") as response:
                server_time = (await response.json())["serverTime"] / 1000
                self.rate_limiter.time_offset = server_time - time()

    async def _get_chunk_periods(
        self, start: int, end: int, interval: str
    ) -> List[Tuple[int, int]]:
        chunk_size = 1000  # Maximum number of candles per request
        interval_ms = {
            "1m": 60000,
            "3m": 180000,
            "5m": 300000,
            "15m": 900000,
            "30m": 1800000,
            "1h": 3600000,
            "2h": 7200000,
            "4h": 14400000,
            "6h": 21600000,
            "8h": 28800000,
            "12h": 43200000,
            "1d": 86400000,
            "3d": 259200000,
            "1w": 604800000,
            "1M": 2592000000,
        }

        step = chunk_size * interval_ms[interval]
        chunks = []
        chunk_start = start

        while chunk_start < end:
            chunk_end = min(chunk_start + step, end)
            chunks.append((chunk_start, chunk_end))
            chunk_start = chunk_end + 1

        return chunks

    async def _fetch_ohlcv_chunk(
        self,
        session: aiohttp.ClientSession,
        worker_id: str,
        symbol: str,
        interval: str,
        start: int,
        end: int,
    ) -> List[List[Any]]:

        params = {
            "symbol": symbol,
            "interval": interval,
            "startTime": start,
            "endTime": end,
            "limit": KLINES_LIMIT,
        }

        for attempt in range(self.max_retries):
            try:
                async with self.semaphore:
                    # check with the Rate Limit Manager if we need to wait
                    # because we exceeded the API rate limit or got banned
                    if wait_time := self.rate_limit_manager.get_wait_time() > 0:
                        logger.warning(
                            "[%s] Rate limit in effect. Waiting for %s seconds.",
                            worker_id,
                            round(wait_time),
                        )
                        await asyncio.sleep(wait_time)

                    # check with the CircuitBreaker if we need to wait because
                    # too many erros occured
                    if self.circuit_breaker.is_open():
                        logger.warning(
                            "Circuit breaker is open. Waiting before retrying..."
                        )
                        await asyncio.sleep(self.circuit_breaker.seconds_to_reset())

                    # check with the RateLimiter if we need to throttle
                    # our rate of requests and wait to avoid hitting the
                    # API rate limit(s)
                    delay, count = 0.0, 0
                    if self.rate_limiter.should_throttle():
                        delay, count = self.rate_limiter.chill_out_bro()
                        logger.info(
                            "[%s] throttling %s: %s",
                            worker_id,
                            count,
                            round(delay, 2),
                        )
                        await asyncio.sleep(delay)

                    # send the downlaod request to the API
                    async with session.get(
                        f"{BASE_URL}{API_ENDPOINT}",
                        params=params,
                    ) as response:
                        self.rate_limiter.update(
                            int(response.headers.get("x-mbx-used-weight-1m", 0)),
                            int(response.headers.get("x-mbx-used-weight", 0)),
                        )

                        # simulates 429 errors (too many requests) if the
                        # SIMULATE_429_ERROS switch has been set (at top of file)
                        if SIMULATE_429_ERRORS and random.random() > 0.99:
                            response.status = 429

                        # handle a 429 error (too many requests)
                        if response.status == 429:
                            logger.warning("[%s] Hit a 429 error!" % worker_id)

                            # get the seconds to wait for a retry from
                            # the headers of the response, or set it to
                            # 10 if we are only simulating a 429
                            if SIMULATE_429_ERRORS:
                                retry_after = 10
                            else:
                                retry_after = int(
                                    response.headers.get("Retry-After", 60)
                                )

                            self.rate_limit_manager.set_rate_limit(retry_after)
                            return await self._fetch_ohlcv_chunk(
                                session,
                                worker_id,
                                symbol,
                                interval,
                                start,
                                end,
                            )

                        # handle a 418 error (= banned)
                        if response.status == 418:
                            retry_after = int(response.headers.get("Retry-After", 120))
                            logger.error(
                                "[%s] We are BANNED for %s seconds"
                                % (worker_id, retry_after)
                            )
                            self.rate_limit_manager.set_ban(retry_after)
                            return await self._fetch_ohlcv_chunk(
                                session,
                                worker_id,
                                symbol,
                                interval,
                                start,
                                end,
                            )

                        response.raise_for_status()

                        logger.info(
                            "[%s] Request successful. Weight: %s-%s",
                            worker_id,
                            self.rate_limiter.weight_1m,
                            self.rate_limiter.weight_total,
                        )

                        return await response.json()

            except asyncio.TimeoutError:
                logger.warning("[%s] Request timed out. Retrying..." % worker_id)
                self.circuit_breaker.record_failure()
                if attempt == self.max_retries - 1:
                    raise

            except (aiohttp.ClientResponseError, Exception) as e:
                logger.error(
                    "[%s] Error: %s for %s. Retrying in %s seconds...",
                    worker_id,
                    e,
                    symbol,
                    f"{wait_time:.2f}",
                )
                self.circuit_breaker.record_failure()
                if attempt == self.max_retries - 1:
                    raise
                wait_time = (2**attempt) + random.uniform(0, 1)
                await asyncio.sleep(wait_time)

        logger.warning(
            "no result for %s after retrying %s times" % (symbol, self.max_retries)
        )
        return []

    async def get_ohlcv(
        self, symbol: str, interval: str, start: int, end: int
    ) -> List[List[Any]]:
        """
        Asynchronously fetches OHLCV (Open, High, Low, Close, Volume)
        data for a given symbol and time range.

        This method breaks down the requested time range into chunks,
        fetches data for each chunk in parallel,filters out any errors,
        and returns the combined and sorted results.

        Arguments:
        ----------
        symbol: str
            The trading pair symbol (e.g., 'BTCUSDT').
        interval: str
            The candlestick interval (e.g., '1m', '1h', '1d').
        start:int
            The start time in milliseconds since the Unix epoch.
        end: int
            The end time in milliseconds since the Unix epoch.

        Returns:
        --------
        List: [List[Any]]
            A list of OHLCV data points, where each data point is a list
            containing [timestamp, open, high, low, close, volume, ...]
            in that order. The list is sorted by timestamp in ascending
            order.
        """
        chunks = await self._get_chunk_periods(start, end, interval)

        async with aiohttp.ClientSession(timeout=self.timeout) as session:
            tasks = [
                self._fetch_ohlcv_chunk(
                    session,
                    f"{symbol}_{i}",
                    symbol,
                    interval,
                    chunk_start,
                    chunk_end,
                )
                for i, (chunk_start, chunk_end) in enumerate(chunks)
            ]
            results = await asyncio.gather(*tasks)

        # Filter out any exceptions and log them
        filtered_results = []
        for i, result in enumerate(results):
            if isinstance(result, Exception):
                logger.error(f"Error in chunk {i} for {symbol}: {result}")
            else:
                filtered_results.extend(result)

        # Sort the results by open time (first element in each candle)
        sorted_results = sorted(filtered_results, key=lambda x: x[0])

        return sorted_results

    async def download_ohlcv_data(
        self, symbols: List[str], interval: str, start: int, end: int
    ) -> Dict[str, List[List[Any]]]:
        """
        Asynchronously download OHLCV (Open, High, Low, Close, Volume)
        data for multiple symbols.

        This function creates concurrent tasks to fetch OHLCV data for
        each symbol in the provided list, over the specified time range
        and interval.

        Parameters:
        -----------
        symbols : List[str]
            A list of trading pair symbols (e.g., ['BTCUSDT', 'ETHUSDT']).
        interval : str
            The candlestick interval (e.g., '1m', '1h', '1d').
        start : int
            The start time in milliseconds since the Unix epoch.
        end : int
            The end time in milliseconds since the Unix epoch.

        Returns:
        --------
        Dict[str, List[List[Any]]]
            A dictionary where each key is a symbol and the corresponding
            value is a list of OHLCV data points. Each data point is a
            list containing [timestamp, open, high, low, close, volume, ...]
            in that order.
        """
        tasks = [self.get_ohlcv(symbol, interval, start, end) for symbol in symbols]
        results = await asyncio.gather(*tasks)
        return dict(zip(symbols, results))


# Usage
async def main():
    client = BinanceClient()
    symbols = ["BTCUSDT", "ETHUSDT", "BNBUSDT", "XRPUSDT", "XLMUSDT"]
    interval = "15m"
    start = int(time() * 1000) - 5 * 365 * 24 * 60 * 60 * 1000  # 5 years ago
    end = int(time() * 1000)

    st = time()

    data = await client.download_ohlcv_data(symbols, interval, start, end)

    et = int((time() - st) * 1000)
    no_of_klines = sum(len(v) for v in data.values())
    print(f"got {no_of_klines} klines in {et}ms")

    for symbol, ohlcv in data.items():
        print(f"{symbol}: {len(ohlcv)} candles downloaded")


if __name__ == "__main__":
    asyncio.run(main())



# -*- coding: utf-8 -*-

# built ins
import asyncio

# installed
import orjson
from loguru import logger as log

# user defined formula
from streaming_helper.restful_api.deribit import end_point_params_template
from streaming_helper.restful_api import connector
from streaming_helper.db_management import redis_client, sqlite_management as db_mgt
from streaming_helper.transaction_management.deribit import (
    cancelling_active_orders as cancel_order,
)
from streaming_helper.channel_management.deribit import subscribing_to_channels
from streaming_helper.strategies.deribit import basic_strategy
from streaming_helper.utilities import (
    string_modification as str_mod,
    caching,
    system_tools as tools,
    error_handling,
)


async def processing_orders(
    client_id: str,
    client_secret: str,
    client_redis: object,
    cancellable_strategies: list,
    currencies: list,
    initial_data_subaccount: dict,
    order_db_table: str,
    redis_channels: list,
    strategy_attributes: list,
) -> None:
    """ """

    try:

        # connecting to redis pubsub
        pubsub: object = client_redis.pubsub()

        # subscribe to channels
        await subscribing_to_channels.redis_channels(
            pubsub,
            redis_channels,
            "processing_orders",
        )

        strategy_attributes_active = [
            o for o in strategy_attributes if o["is_active"] == True
        ]

        # get strategies that have not short/long attributes in the label
        non_checked_strategies = [
            o["strategy_label"]
            for o in strategy_attributes_active
            if o["non_checked_for_size_label_consistency"] == True
        ]

        # get redis channels
        order_rest_channel: str = redis_channels["order_rest"]
        my_trade_receiving_channel: str = redis_channels["my_trade_receiving"]
        order_update_channel: str = redis_channels["order_cache_updating"]
        portfolio_channel: str = redis_channels["portfolio"]
        sqlite_updating_channel: str = redis_channels["sqlite_record_updating"]
        sub_account_cached_channel: str = redis_channels["sub_account_cache_updating"]

        not_cancel = True

        query_trades = f"SELECT * FROM  v_trading_all_active"

        sub_account_cached_params = initial_data_subaccount["params"]

        sub_account_cached = sub_account_cached_params["data"]

        orders_cached = sub_account_cached["orders_cached"]

        positions_cached = sub_account_cached["positions_cached"]

        ordered = []

        connection_url = end_point_params_template.basic_https()

        subaccounts_details_end_point = (
            end_point_params_template.get_subaccounts_details_end_point()
        )

        while not_cancel:

            try:

                message_byte = await pubsub.get_message()

                if message_byte and message_byte["type"] == "message":

                    message_byte_data = orjson.loads(message_byte["data"])

                    params = message_byte_data["params"]

                    data = params["data"]

                    message_channel = params["channel"]

                    log.warning(message_channel)

                    if my_trade_receiving_channel in message_channel:

                        log.critical(message_channel)
                        log.error(data)

                        for trade in data:

                            currency_lower: str = trade["fee_currency"].lower()

                            archive_db_table = f"my_trades_all_{currency_lower}_json"

                            await cancel_order.cancel_the_cancellables(
                                client_id,
                                client_secret,
                                order_db_table,
                                currency_lower,
                                cancellable_strategies,
                            )

                            await saving_traded_orders(
                                trade,
                                archive_db_table,
                                order_db_table,
                            )

                            for currency in currencies:

                                subaccounts_details_params = end_point_params_template.get_subaccounts_details_params(
                                    currency
                                )

                                result = await connector.get_connected(
                                    connection_url,
                                    subaccounts_details_end_point,
                                    client_id,
                                    client_secret,
                                    subaccounts_details_params,
                                )

                                await updating_sub_account(
                                    client_redis,
                                    orders_cached,
                                    positions_cached,
                                    query_trades,
                                    result["result"],
                                    sub_account_cached_channel,
                                    message_byte_data,
                                )

                    if order_rest_channel in message_channel:

                        log.critical(message_channel)
                        log.warning(data)

                        if data["order_allowed"]:

                            await if_order_is_true(
                                client_id,
                                client_secret,
                                connection_url,
                                non_checked_strategies,
                                data,
                                ordered,
                            )

                    if order_update_channel in message_channel:

                        data = data["current_order"]

                        log.critical(message_channel)
                        log.warning(data)

                        if "oto_order_ids" in data:

                            await saving_oto_order(
                                client_id,
                                client_secret,
                                non_checked_strategies,
                                [data],
                                order_db_table,
                                ordered,
                            )

                        else:

                            if "OTO" not in data["order_id"]:

                                label = data["label"]

                                order_id = data["order_id"]
                                order_state = data["order_state"]
                             
                                # no label
                                if label == "" or label == '':

                                    await cancelling_and_relabelling(
                                        client_id,
                                        client_secret,
                                        non_checked_strategies,
                                        order_db_table,
                                        data,
                                        label,
                                        order_state,
                                        order_id,
                                        ordered,
                                    )

                                else:
                                    label_and_side_consistent = (
                                        basic_strategy.is_label_and_side_consistent(
                                            non_checked_strategies, data
                                        )
                                    )

                                    if label_and_side_consistent and label:

                                        # log.debug(f"order {order}")

                                        await saving_order_based_on_state(
                                            order_db_table,
                                            data,
                                        )

                                    # check if transaction has label. Provide one if not any
                                    if not label_and_side_consistent:

                                        if (
                                            order_state != "cancelled"
                                            or order_state != "filled"
                                        ):
                                            await cancel_order.cancel_by_order_id(
                                                client_id,
                                                client_secret,
                                                order_db_table,
                                                order_id,
                                            )

                                            # log.error (f"order {order}")
                                            await db_mgt.insert_tables(
                                                order_db_table,
                                                data,
                                            )

                        for currency in currencies:

                            subaccounts_details_params = end_point_params_template.get_subaccounts_details_params(
                                currency
                            )

                            result = await connector.get_connected(
                                connection_url,
                                subaccounts_details_end_point,
                                client_id,
                                client_secret,
                                subaccounts_details_params,
                            )

                            await updating_sub_account(
                                client_redis,
                                orders_cached,
                                positions_cached,
                                query_trades,
                                result["result"],
                                sub_account_cached_channel,
                                message_byte_data,
                            )

                    if (
                        sqlite_updating_channel in message_channel
                        or portfolio_channel in message_channel
                    ):
                        for currency in currencies:

                            subaccounts_details_params = end_point_params_template.get_subaccounts_details_params(
                                currency
                            )

                            result = await connector.get_connected(
                                connection_url,
                                subaccounts_details_end_point,
                                client_id,
                                client_secret,
                                subaccounts_details_params,
                            )

                            await updating_sub_account(
                                client_redis,
                                orders_cached,
                                positions_cached,
                                query_trades,
                                result["result"],
                                sub_account_cached_channel,
                                message_byte_data,
                            )

            except Exception as error:

                await error_handling.parse_error_message_with_redis(
                    client_redis,
                    error,
                )

                continue

            finally:
                await asyncio.sleep(0.001)

    except Exception as error:

        await error_handling.parse_error_message_with_redis(
            client_redis,
            error,
        )


async def if_order_is_true(
    client_id: str,
    client_secret: str,
    non_checked_strategies: list,
    order: dict,
    ordered: list,
) -> None:
    """ """

    if order["order_allowed"]:

        # get parameter orders
        try:
            params: dict = order["order_parameters"]
        except:
            params: dict = order

        label_and_side_consistent: bool = basic_strategy.is_label_and_side_consistent(
            non_checked_strategies,
            params,
        )

        if label_and_side_consistent:

            label = str_mod.parsing_label(params["label"])["main"]
            log.warning(f"params {params}")
            log.debug(f"label {label}")
            label_has_order_id = [o["order_id"] for o in ordered if label in o["label"]]
            log.debug(f"label_has_order_id {label_has_order_id}")

            if label_has_order_id:
                ordered.remove(params)

            else:
                if ordered == []:

                    send_limit_order_end_point = (
                        end_point_params_template.send_orders_end_point(params)
                    )

                    send_limit_order_params = (
                        end_point_params_template.send_limit_order_params(params)
                    )

                    connection_url = end_point_params_template.basic_https()

                    send_limit_result = await connector.get_connected(
                        connection_url,
                        send_limit_order_end_point,
                        client_id,
                        client_secret,
                        send_limit_order_params,
                    )

                    return send_limit_result

                else:
                    ordered.append(params)

                    return []

            # await asyncio.sleep(10)
        else:

            return []
            # await asyncio.sleep(10)


async def cancelling_and_relabelling(
    client_id: str,
    client_secret: str,
    non_checked_strategies,
    order_db_table,
    order,
    label,
    order_state,
    order_id,
    ordered,
) -> None:

    log.debug(f"label {label} order_state {order_state} order {order}")
    
    log.debug(order["order_id"])
    
    log.error(label == "" or label == '')

    # no label
    if label == "" or label == '':

        # log.info(label == "")

        if "open" in order_state or "untriggered" in order_state:
            
            log.error("OTO" not in order["order_id"])

            if "OTO" not in order["order_id"]:

                # log.error (f"order {order}")
                await db_mgt.insert_tables(
                    order_db_table,
                    order,
                )

                await cancel_order.cancel_by_order_id(
                    client_id,
                    client_secret,
                    order_db_table,
                    order_id,
                )

            order_attributes = labelling_unlabelled_order(order)
            # log.warning (f"order_attributes {order_attributes}")

            await if_order_is_true(
                client_id,
                client_secret,
                non_checked_strategies,
                order_attributes,
                ordered,
            )


def labelling_unlabelled_order(order: dict) -> None:

    type = order["order_type"]

    side = basic_strategy.get_transaction_side(order)
    order.update({"everything_is_consistent": True})
    order.update({"order_allowed": True})
    order.update({"entry_price": order["price"]})
    order.update({"size": order["amount"]})
    order.update({"type": type})

    if type != "limit":  # limit has various state
        order.update({"trigger_price": order["trigger_price"]})
        order.update({"trigger": order["trigger"]})

    order.update({"side": side})

    label_open: str = get_custom_label(order)
    order.update({"label": label_open})

    return order


def labelling_unlabelled_order_oto(
    transaction_main: list,
    transaction_secondary: list,
) -> None:
    """

    orders_example= [
        {
            'oto_order_ids': ['OTO-80322590'], 'is_liquidation': False, 'risk_reducing': False,
            'order_type': 'limit', 'creation_timestamp': 1733172624209, 'order_state': 'open',
            'reject_post_only': False, 'contracts': 1.0, 'average_price': 0.0, 'reduce_only': False,
            'trigger_fill_condition': 'incremental', 'last_update_timestamp': 1733172624209,
            'filled_amount': 0.0, 'replaced': False, 'post_only': True, 'mmp': False, 'web': True,
            'api': False, 'instrument_name': 'BTC-PERPETUAL', 'max_show': 10.0, 'time_in_force': 'good_til_cancelled',
            'direction': 'buy', 'amount': 10.0, 'order_id': '81944428472', 'price': 90000.0, 'label': ''},
        {
            'is_liquidation': False, 'risk_reducing': False, 'order_type': 'limit',
            'creation_timestamp': 1733172624177, 'order_state': 'untriggered', 'average_price': 0.0,
            'reduce_only': False, 'trigger_fill_condition': 'incremental', 'last_update_timestamp': 1733172624177,
            'filled_amount': 0.0, 'is_secondary_oto': True, 'replaced': False, 'post_only': False, 'mmp': False,
            'web': True, 'api': False, 'instrument_name': 'BTC-PERPETUAL', 'max_show': 10.0,
            'time_in_force': 'good_til_cancelled', 'direction': 'sell', 'amount': 10.0,
            'order_id': 'OTO-80322590', 'price': 100000.0, 'label': ''}
        ]


    """

    label: str = get_custom_label_oto(transaction_main)

    label_open: str = label["open"]
    label_closed: str = label["closed"]
    instrument_name = transaction_main["instrument_name"]
    secondary_params = [
        {
            "amount": transaction_secondary["amount"],
            "direction": (transaction_secondary["direction"]),
            "type": "limit",
            "instrument_name": instrument_name,
            "label": label_closed,
            "price": transaction_secondary["price"],
            "time_in_force": "good_til_cancelled",
            "post_only": True,
        }
    ]

    params = {}
    params.update({"everything_is_consistent": True})
    params.update({"instrument_name": instrument_name})
    params.update({"type": "limit"})
    params.update({"entry_price": transaction_main["price"]})
    params.update({"size": transaction_main["amount"]})
    params.update({"label": label_open})
    params.update({"side": transaction_main["direction"]})
    params.update({"otoco_config": secondary_params})

    try:
        params.update({"linked_order_type": transaction_main["linked_order_type"]})

    except:
        pass  # default: one_triggers_other at API request

    return dict(
        order_allowed=True,
        order_parameters=params,
    )


def get_custom_label_oto(transaction: list) -> dict:

    side = transaction["direction"]
    side_label = "Short" if side == "sell" else "Long"

    try:
        last_update = transaction["timestamp"]
    except:
        try:
            last_update = transaction["last_update_timestamp"]
        except:
            last_update = transaction["creation_timestamp"]

    return dict(
        open=(f"custom{side_label.title()}-open-{last_update}"),
        closed=(f"custom{side_label.title()}-closed-{last_update}"),
    )


def get_custom_label(transaction: list) -> str:

    side = transaction["direction"]
    side_label = "Short" if side == "sell" else "Long"

    try:
        last_update = transaction["timestamp"]
    except:
        try:
            last_update = transaction["last_update_timestamp"]
        except:
            last_update = transaction["creation_timestamp"]

    return f"custom{side_label.title()}-open-{last_update}"


async def saving_order_based_on_state(
    order_table: str,
    order: dict,
    db: str = "sqlite",
) -> None:
    """
    db: "sqlite"
        "redis"

    _summary_

    Args:
        trades (_type_): _description_
        orders (_type_): _description_

    Examples:

        original=  {
            'jsonrpc': '2.0',
            'id': 1002,
            'result': {
                'trades': [],
                'order': {
                    'is_liquidation': False, 'risk_reducing': False, 'order_type': 'limit',
                    'creation_timestamp': 1728090482863, 'order_state': 'open', 'reject_post_only': False,
                    'contracts': 5.0, 'average_price': 0.0, 'reduce_only': False,
                    'last_update_timestamp': 1728090482863, 'filled_amount': 0.0, 'post_only': True,
                    'replaced': False, 'mmp': False, 'order_id': 'ETH-49960097702', 'web': False, 'api': True,
                    'instrument_name': 'ETH-PERPETUAL', 'max_show': 5.0, 'time_in_force': 'good_til_cancelled',
                    'direction': 'sell', 'amount': 5.0, 'price': 2424.05, 'label': 'hedgingSpot-open-1728090482812'
                    }
                    },
                    'usIn': 1728090482862653,
                    'usOut': 1728090482864640,
                    'usDiff': 1987,
                    'testnet': False
                    }

        cancelled=  {
            'jsonrpc': '2.0',
            'id': 1002,
            'result': {
                'is_liquidation': False, 'risk_reducing': False, 'order_type': 'limit',
                'creation_timestamp': 1728090482863, 'order_state': 'cancelled','reject_post_only': False,
                'contracts': 5.0, 'average_price': 0.0,'reduce_only': False,
                'last_update_timestamp': 1728090483773, 'filled_amount': 0.0,'post_only': True,
                'replaced': False, 'mmp': False, 'cancel_reason': 'user_request','order_id': 'ETH-49960097702',
                'web': False, 'api': True, 'instrument_name': 'ETH-PERPETUAL',
                'max_show': 5.0, 'time_in_force': 'good_til_cancelled', 'direction': 'sell', 'amount': 5.0,
                'price': 2424.05, 'label': 'hedgingSpot-open-1728090482812'},
                'usIn': 1728090483773107,
                'usOut': 1728090483774372,
                'usDiff': 1265,
                'testnet': False
                }

    """

    filter_trade = "order_id"

    order_id = order[f"{filter_trade}"]

    order_state = order["order_state"]

    if order_state == "cancelled" or order_state == "filled":

        await db_mgt.deleting_row(
            order_table,
            "databases/trading.sqlite3",
            filter_trade,
            "=",
            order_id,
        )

    if order_state == "open":

        # log.error (f"order {order}")
        await db_mgt.insert_tables(
            order_table,
            order,
        )


async def saving_traded_orders(
    trade_result: str,
    trade_table: str,
    order_db_table: str,
) -> None:
    """_summary_

    Args:
        trades (_type_): _description_
        orders (_type_): _description_
    """

    filter_trade = "order_id"

    order_id = trade_result[f"{filter_trade}"]

    # remove respective transaction from order db
    await db_mgt.deleting_row(
        order_db_table,
        "databases/trading.sqlite3",
        filter_trade,
        "=",
        order_id,
    )

    trade_to_db = trade_template()

    trade_to_db.update({"instrument_name": trade_result["instrument_name"]})
    trade_to_db.update({"amount": trade_result["amount"]})
    trade_to_db.update({"price": trade_result["price"]})
    trade_to_db.update({"direction": trade_result["direction"]})
    trade_to_db.update({"trade_id": trade_result["trade_id"]})
    trade_to_db.update({"order_id": trade_result["order_id"]})
    trade_to_db.update({"timestamp": trade_result["timestamp"]})

    try:
        trade_to_db.update({"label": trade_result["label"]})

    except:

        pass

    await db_mgt.insert_tables(
        trade_table,
        trade_to_db,
    )


async def saving_oto_order(
    client_id: str,
    client_secret: str,
    non_checked_strategies: list,
    orders: list,
    order_db_table: str,
    ordered: list,
) -> None:

    print(f"saving_oto_order {orders}")
    transaction_main = [o for o in orders if "OTO" not in o["order_id"]][0]

    transaction_main_oto = transaction_main["oto_order_ids"][0]

    kind = "future"
    type = "trigger_all"

    open_orders_end_point = end_point_params_template.get_open_orders_end_point()

    open_orders_params = end_point_params_template.send_limit_order_params(kind, type)

    connection_url = end_point_params_template.basic_https()

    open_orders = await connector.get_connected(
        connection_url,
        open_orders_end_point,
        client_id,
        client_secret,
        open_orders_params,
    )

    open_orders_from_exchange = open_orders["result"]

    transaction_secondary = [
        o for o in open_orders_from_exchange if transaction_main_oto in o["order_id"]
    ]

    if transaction_secondary:

        transaction_secondary = transaction_secondary[0]

        # no label
        if (
            transaction_main["label"] == ""
            and "open" in transaction_main["order_state"]
        ):

            # log.error (f"transaction_main {transaction_main}")
            await db_mgt.insert_tables(
                order_db_table,
                transaction_main,
            )

            order_attributes = labelling_unlabelled_order_oto(
                transaction_main, transaction_secondary
            )

            await cancel_order.cancel_by_order_id(
                client_id,
                client_secret,
                order_db_table,
                transaction_main["order_id"],
            )

            await if_order_is_true(
                client_id,
                client_secret,
                non_checked_strategies,
                order_attributes,
                ordered,
            )

        else:

            # log.error (f"transaction_main {transaction_main}")
            await db_mgt.insert_tables(
                order_db_table,
                transaction_main,
            )


async def updating_sub_account(
    client_redis: object,
    orders_cached: list,
    positions_cached: list,
    query_trades: str,
    subaccounts_details_result: list,
    sub_account_cached_channel: str,
    message_byte_data: dict,
) -> None:

    if subaccounts_details_result:

        open_orders = [o["open_orders"] for o in subaccounts_details_result]

        if open_orders:
            caching.update_cached_orders(
                orders_cached,
                open_orders[0],
                "rest",
            )
        positions = [o["positions"] for o in subaccounts_details_result]

        if positions:
            caching.positions_updating_cached(
                positions_cached,
                positions[0],
                "rest",
            )

    my_trades_active_all = await db_mgt.executing_query_with_return(query_trades)

    data = dict(
        positions=positions_cached,
        open_orders=orders_cached,
        my_trades=my_trades_active_all,
    )

    message_byte_data["params"].update({"channel": sub_account_cached_channel})
    message_byte_data["params"].update({"data": data})

    await redis_client.publishing_result(
        client_redis,
        message_byte_data,
    )


def is_order_has_executed(
    ordered: list,
    orders_from_exchange: list,
) -> bool:

    order_has_executed = [
        o for o in ordered if o["label"] in [o["label"] for o in orders_from_exchange]
    ]

    if order_has_executed:
        ordered = []

    return True if order_has_executed else False


def trade_template() -> str:
    """
    combining result from websocket (user changes/trade)
        and rest API (get_transaction_log)

        instrument_name: str=None,
        amount: int=None,
        price: float=None,
        side: str=None,
        direction: str=None,
        position: int=None,
        currency: str=None,
        timestamp: int=None,
        trade_id: str=None,
        order_id: str=None,
        user_seq: int=None

    label is not registered since it needed as identification for no label
    """
    return dict(
        instrument_name=None,
        amount=None,
        price=None,
        side=None,
        direction=None,
        position=None,
        currency=None,
        timestamp=None,
        trade_id=None,
        order_id=None,
        user_seq=None,
    )



# -*- coding: utf-8 -*-

# built ins
import asyncio

# installed
import uvloop
from dataclassy import dataclass
import orjson

import redis.asyncio as aioredis

asyncio.set_event_loop_policy(uvloop.EventLoopPolicy())

from streaming_helper.utilities import error_handling


class RedisPubSubManager:
    """

    https://medium.com/@nandagopal05/scaling-websockets-with-pub-sub-using-python-redis-fastapi-b16392ffe291
        Initializes the RedisPubSubManager.

    Args:
        host (str): Redis server host.
        port (int): Redis server port.
    """

    def __init__(self, host="localhost", port=6379):
        self.redis_host = host
        self.redis_port = port
        self.pubsub = None

    async def _get_redis_connection(self) -> aioredis.Redis:
        """
        Establishes a connection to Redis.

        Returns:
            aioredis.Redis: Redis connection object.
        """
        return aioredis.Redis(
            host=self.redis_host, port=self.redis_port, auto_close_connection_pool=False
        )

    async def connect(self) -> None:
        """
        Connects to the Redis server and initializes the pubsub client.
        """
        self.redis_connection = await self._get_redis_connection()
        self.pubsub = self.redis_connection.pubsub()

    async def _publish(self, room_id: str, message: str) -> None:
        """
        Publishes a message to a specific Redis channel.

        Args:
            room_id (str): Channel or room ID.
            message (str): Message to be published.
        """
        await self.redis_connection.publish(room_id, message)

    async def subscribe(self, room_id: str) -> aioredis.Redis:
        """
        Subscribes to a Redis channel.

        Args:
            room_id (str): Channel or room ID to subscribe to.

        Returns:
            aioredis.ChannelSubscribe: PubSub object for the subscribed channel.
        """
        await self.pubsub.subscribe(room_id)
        return self.pubsub

    async def unsubscribe(self, room_id: str) -> None:
        """
        Unsubscribes from a Redis channel.

        Args:
            room_id (str): Channel or room ID to unsubscribe from.
        """
        await self.pubsub.unsubscribe(room_id)


class Singleton(type):
    """
    https://stackoverflow.com/questions/49398590/correct-way-of-using-redis-connection-pool-in-python
    """

    _instances = {}

    def __call__(cls, *args, **kwargs):
        if cls not in cls._instances:
            cls._instances[cls] = super(Singleton, cls).__call__(*args, **kwargs)
        return cls._instances[cls]


@dataclass(unsafe_hash=True, slots=True)
class RedisClient(metaclass=Singleton):

    host: str = "redis://localhost"
    port: int = 6379
    db: int = 0
    protocol: int = 3
    pool: object = None

    def __post_init__(self):
        return redis.Redis.from_pool(
            redis.ConnectionPool.from_url(
                self.host,
                port=self.port,
                db=self.db,
                protocol=self.protocol,
            )
        )

    def conn(self):
        return self.pool


async def saving_and_publishing_result(
    client_redis: object,
    channel: str,
    keys: str,
    cached_data: list,
    message: dict,
) -> None:
    """ """

    try:
        # updating cached data
        if cached_data:
            await saving_result(
                client_redis,
                channel,
                keys,
                cached_data,
            )

        # publishing message
        await publishing_result(
            client_redis,
            channel,
            message,
        )

    except Exception as error:

        await error_handling.parse_error_message_with_redis(
            client_redis,
            error,
        )


async def publishing_result(
    client_redis: object,
    message: dict,
) -> None:
    """ """

    try:

        channel = message["params"]["channel"]

        # publishing message
        await client_redis.publish(
            channel,
            orjson.dumps(message),
        )

    except Exception as error:

        await error_handling.parse_error_message_with_redis(
            client_redis,
            error,
        )


async def saving_result(
    client_redis: object,
    channel: str,
    keys: str,
    cached_data: list,
) -> None:
    """ """

    try:

        channel = message["channel"]

        await client_redis.hset(
            keys,
            channel,
            orjson.dumps(cached_data),
        )

    except Exception as error:

        await error_handling.parse_error_message_with_redis(
            client_redis,
            error,
        )


async def querying_data(
    client_redis: object,
    channel: str,
    keys: str,
) -> None:
    """ """

    try:

        return await client_redis.hget(
            keys,
            channel,
        )

    except Exception as error:

        await error_handling.parse_error_message_with_redis(
            client_redis,
            error,
        )


async def publishing_specific_purposes(
    purpose,
    message,
    redis_channels: list = None,
    client_redis: object = None,
) -> None:
    """
    purposes:
    + porfolio
    + sub_account_update
    + trading_update

        my_trades_channel:
        + send messages that "high probabilities" trade DB has changed
            sender: redis publisher + sqlite insert, update & delete
        + updating trading cache at end user
            consumer: fut spread, hedging, cancelling
        + checking data integrity
            consumer: app data cleaning/size reconciliation

        sub_account_channel:
        + send messages that sub_account has changed
            sender: deribit API module
        + updating sub account cache at end user
            consumer: fut spread, hedging, cancelling
        + checking data integrity
            consumer: app data cleaning/size reconciliation

    """

    try:

        if not client_redis:
            pool = aioredis.ConnectionPool.from_url(
                "redis://localhost",
                port=6379,
                db=0,
                protocol=3,
                decode_responses=True,
            )
            client_redis: object = aioredis.Redis.from_pool(pool)

        if not redis_channels:

            from streaming_helper.utilities.system_tools import get_config_tomli

            # registering strategy config file
            file_toml = "config_strategies.toml"

            # parsing config file
            config_app = get_config_tomli(file_toml)

            # get redis channels
            redis_channels: dict = config_app["redis_channels"][0]

        if purpose == "sqlite_record_updating":
            channel: str = redis_channels["sqlite_record_updating"]
            message["params"].update({"channel": channel})

        await publishing_result(
            client_redis,
            message,
        )

    except Exception as error:

        await error_handling.parse_error_message_with_redis(
            client_redis,
            error,
        )
